#!/usr/bin/env python
# -*- coding: UTF-8 -*-
# --------------------------------------------------------------------------
# Get YouTube Video, Copyright (C) 2006-2011 Otto Allmendinger, Hirs, Mathias Panzenböck
# --------------------------------------------------------------------------
# ***** BEGIN GPL LICENSE BLOCK *****
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
#
# ***** END GPL LICENCE BLOCK *****
# --------------------------------------------------------------------------

# version: 1.10
#
# differnet versions:
#  by Otto Allmendinger <otto.allmendinger@gmail.com> (original author):
#     http://www.kde-apps.org/content/show.php?content=37509
#  by Hirs:
#     http://www.kde-apps.org/content/show.php?content=41432
#  by panzi (this version):
#     http://www.kde-apps.org/content/show.php?content=41456
#
# further credits:
# spanish translation of the menu entries by Víctor Fernández Martínez
# improvements to video encoding by Tejas Dinkar <tejasdinkar AT gmail DOT com>
# tipps about warnings and errors in python 2.3 by Volker Janzen
# (However, I dropped python 2.3 support now.)
# Russian and Ukrainian translations by Ivan Lezhnev Jr.

import urllib, re, os, sys
import gettext, locale
from subprocess import Popen, PIPE
from htmlentitydefs import name2codepoint
from lxml import html

import traceback
from ConfigParser import ConfigParser
from getopt import gnu_getopt as getopt

if os.getenv('KDE_SESSION_VERSION') == '4':
	FILE_CLIENT='kioclient'
	KCOOKIEJAR='kcookiejar4'
	try:
		import dbus
	except ImportError:
		GET_COOKIES=lambda: cmd('qdbus', 'org.kde.kded', '/modules/kcookiejar',
			'org.kde.KCookieServer.findDOMCookies', 'http://www.youtube.com')
	else:
		_get_cookies = dbus.SessionBus().get_object('org.kde.kded',
			'/modules/kcookiejar').get_dbus_method('findDOMCookies',
				dbus_interface='org.kde.KCookieServer')
		GET_COOKIES=lambda: (_get_cookies('http://www.youtube.com', 0), '', 0)

else:
	FILE_CLIENT='kfmclient'
	KCOOKIEJAR='kcookiejar'
	GET_COOKIES=lambda: cmd('dcop', 'kcookiejar', 'kcookiejar',
		'findCookies(QString)', 'http://www.youtube.com')

def shesc(*s):
	if len(s) == 1:
		return "'%s'" % s[0].replace("'","'\\''")
	else:
		return tuple(("'%s'" % arg.replace("'","'\\''")) for arg in s)

def strconv(f):
	def __strconv(*args,**kwargs):
		return ''.join(f(*args,**kwargs))
	
	return __strconv

@strconv
def htmlesc(s):
	for c in s:
		if c == '<':
			yield "&lt;"
		elif c == '>':
			yield "&gt;"
		elif c == '"':
			yield "&quot;"
		elif c == '&':
			yield "&amp;"
		else:
			yield c

def htmlmap(s):
	if s[0:1] == '#':
		try:
			if s[1:2] == 'x':
				return unichr(int(s[2:],16))
			else:
				return unichr(int(s[1:],10))
		except ValueError:
			pass

	cp = name2codepoint.get(s,None)

	if cp is None:
		return None

	return unichr(cp)

@strconv
def htmlunesc(s):
	i = 0

	while True:
		j = s.find('&',i)

		if j < 0:
			yield s[i:]
			break

		else:
			yield s[i:j]

			i = j
			j = s.find(';',i)

			if j < 0:
				yield '&'
				i += 1

			else:
				tr = htmlmap(s[i+1:j])

				if tr is None:
					yield '&'
					i += 1

				else:
					yield tr
					i = j + 1

def fixendl(s):
	if s[-1:] == '\n':
		return s[:-1]
	else:
		return s

def cmd(*args,**kwargs):
	p = Popen(args,**kwargs)

	if kwargs.get('stderr',None) == PIPE:
		serr = fixendl(p.stderr.read())
	else:
		serr = ''

	if kwargs.get('stdout',None) == PIPE:
		sout = fixendl(p.stdout.read())
	else:
		sout = ''

	return sout, serr, p.wait()

# dialoges:
def showError(caption,msg):
	return cmd('kdialog', '--caption', caption, '--error', msg)

def showSaveFile(caption,file_name):
	return cmd('kdialog', '--caption', caption, '--getsavefilename', file_name, stdout=PIPE)
	
def showWarningYesNoCancel(caption,msg):
	return cmd('kdialog', '--caption', caption, '--warningyesnocancel', msg)

def showPassivePopup(caption,msg,secs=5):
	return cmd('kdialog', '--caption', caption, '--passivepopup', msg, str(secs))

def get_filename(save_dir, video_id, format):
	url = "http://www.youtube.com/watch?v=" + video_id
	
	try:
		ext, fmt, doc = FORMATS[format]
	except KeyError:
		if format.isdigit():
			ext = 'flv'
			fmt = format
		else:
			raise Exception(_('Unknown file format: %(format)s') % {'format': format})

	stream = urlopener.open(url)

	try:
		data = stream.read()
	finally:
		stream.close()

	doc = html.document_fromstring(data)
#	token   = None
	videoid = None
	name    = None
	fmt_stream_map = None
	
	embed = doc.get_element_by_id("movie_player",None)
	if embed is not None and 'flashvars' in embed.attrib:
		flashvars = dict(map(urllib.unquote_plus,v.split("=",1)) for v in embed.attrib['flashvars'].split("&"))
#		token   = flashvars.get('t',None)
		videoid = flashvars.get('video_id',video_id)
		if 'fmt_stream_map' in flashvars:
			fmt_stream_map = [f.split('|') for f in flashvars['fmt_stream_map'].split(',')]
			fmt_stream_map = dict((info[0], info[1:]) for info in fmt_stream_map)

	if fmt_stream_map is None or videoid is None:
		raise Exception(
			_('<p>No video information found on the video\'s Page.<br/>\n'
			'Maybe you have to be logged in to view this video or YouTube has changed it\'s page layout?<br/>\n'
			'If you are already logged in or you don\'t have to be logged in to view this video, '
			'please report this <a href="http://kde-apps.org/content/show.php?content=41456">here</a>.\n'
			'video-page: <a href="%(url)s">%(url)s</a></p>') %
			{'url':url})

	name = doc.get_element_by_id("eow-title",None)
	if name is None:
		name = videoid
	else:
		name = name.text.strip()
	
	vid_info = fmt_stream_map.get(fmt, None)
	if vid_info is None:
		raise Exception(
			_('<p>Video format %(format)s not supported by video "%(name)s".</p>') %
			{'format':format,'name':name})
	
	user = doc.get_element_by_id("watch-username",None)
	if user is not None:
		name += " by " + ''.join(user.xpath(".//text()")).strip()
	
	# kdialog seems to have problems with "
	name = re.sub('[/:]', '', name).replace('"', "'")

	return "%s/%s.%s" % (save_dir,name,ext), vid_info[0]

def get_video(vid_url, file_name, action):
	if cmd(FILE_CLIENT, 'copy', vid_url, file_name)[2]:
		raise Exception(_('Error saving file: %(file)s') % {'file':file_name})
	
	if action:
		fdir, fname = os.path.split(file_name)
		fname, fext = os.path.splitext(fname)
		cmd_vars = {
			'fdir':  shesc(fdir),
			'fname': shesc(fname),
			'fext':  shesc(fext),
			'fpath': shesc(file_name)
		}
		command = action % cmd_vars

		p = Popen(command,shell=True,stderr=PIPE)
		code = p.wait()

		if code:
			raise Exception(
				_("<p>Error executing command:<br/>\n"
				"<pre>%(command)s</pre><br/>\n"
				"Command output:<br/>\n"
				"<pre>%(output)s</pre></p>") %
				{'command':htmlesc(command),'output':htmlesc(p.stderr.read())})

def usage():
	print("""\
Usage: %s [OPTION...] [VIDEO-ID or URL...]

OPTIONS:
	-h, --help             show this help message

	-a, --action=ACTION    execute ACTION after download
	        following referneces may be included in ACTION:
	            %%(fdir)s .... directory to which the file was saved
	            %%(fname)s ... name of the saved file
	            %%(fext)s .... extension of the saved file
	            %%(fpath)s ... full path of the saved file

	-f, --format=FORMAT    file format to download
	        following formats are supported:""" % sys.argv[0])

	for formats, value in _FORMATS:
		print("""\
	            %s:
	                %s
""" % (', '.join(formats), value[2]))

	print("""\
	            <digits>:
	                specify format code directly (unknown)
""")

def main():
	global urlopener, _, _FORMATS, FORMATS

	try:
		conf_file = os.path.join(os.getenv('HOME'),'.get_yt_videorc')

		bin_dir    = os.path.dirname(sys.argv[0])
		prefix     = os.path.dirname(bin_dir)
		locale_dir = os.path.join(prefix,'share','get_yt_video')

		LANG = os.getenv('LANG')
		
		if locale.getlocale() == (None, None) and LANG is not None:
			try:
				lang = gettext.translation('get_yt_video', locale_dir, languages=[LANG])
				lang.install()
			except:
				_ = lambda s: s

		else:
			gettext.install('get_yt_video', locale_dir)

		_FORMATS = [
			(['flv'],                   ('flv',  '5',  _('Normal 240p (flv)'))),
			(['flv-hq-'],               ('flv',  '6',  _('Old High Quality 480p (flv)'))),
			(['3gp'],                   ('3gp',  '13', _('Mobil Version (3gp)'))),
			(['3gp-hq', '3gp-144p'],    ('3gp',  '17', _('Mobil Version 144p (3gp)'))),
			(['mp4-hq', 'mp4-480p'],    ('mp4',  '18', _('High Quality 480p (mp4)'))),
			(['mp4-hd', 'mp4-720p'],    ('mp4',  '22', _('High Definition 720p (mp4)'))),
			(['mp4-360p'],              ('mp4',  '34', _('High Quality 360p (mp4)'))),
			(['flv-hq', 'flv-480p'],    ('flv',  '35', _('High Quality 480p (flv)'))),
			(['mp4-hd+', 'mp4-1080p'],  ('mp4',  '37', _('High Definition 1080p (mp4)'))),
			(['mp4-hd++', 'mp4-3072p'], ('mp4',  '38', _('High Definition 3072p (mp4)'))),
			(['webm', 'webm-360p'],     ('webm', '43', _('Normal 360p (webm)'))),
			(['webm-720p'],             ('webm', '45', _('High Definition 720p (webm)'))),
		]
		
		FORMATS = {}
		
		for keys, value in _FORMATS:
			keys.append(value[1])
			for key in keys:
				FORMATS[key] = value

		# parse command line options:
		try:
			opts, args = getopt(sys.argv[1:], "ha:f:",
				["help", "action=", "format="])
		except Exception, e:
			sys.stderr.write('*** %s\n' % e)
			sys.exit(2)
		
		action = ''
		format = 'mp4-hq'

		for o, a in opts:
			if o in ("-a", "--action"):
				action = a
			elif o in ("-f", "--format"):
				format = a.strip().lower()
			elif o in ("-h", "--help"):
				usage()
				return
			else:
				assert False, "unhandled option"
		
		if len(args) < 1:
			sys.stderr.write('*** no url given\n')
			usage()
			sys.exit(2)

		# use a FancyURLopener so we can use konquerors cookies :)
		urlopener = urllib.FancyURLopener()
		
		cookies, serr, code = GET_COOKIES()

		if code != 0:
			# try to start kcookiejar
			print('could not get cookies, trying to start cookie service')
			ignore, serr, code = cmd(KCOOKIEJAR)

			if code == 0:
				cookies, serr, code = GET_COOKIES()
				if code != 0 or not cookies:
					if serr:
						print(serr)
					print('could not get cookies (maybe no cookies set yet)')
			else:
				if serr:
					print(serr)
				print('failed to start cookie service')

		if code == 0 and cookies:
			urlopener.addheader(cookies)

		conf = ConfigParser()

		video_ids = []
		
		for url in args:
			# searching for video id in url:
			m = re.search('youtube\.com/v/(?P<video_id>[-A-Za-z0-9_]+)', url)
			
			if m is None:
				m = re.search('video_id=(?P<video_id>[-A-Za-z0-9_]+)', url)
				
			if m is None:
				m = re.search('v=(?P<video_id>[-A-Za-z0-9_]+)', url)
			
			if m is not None:
				# video id found
				video_id = m.group('video_id')
				if video_id not in video_ids:
					video_ids.append(video_id)
	
			else:
				# searching for video ids in content:
				stream = urlopener.open(url)
				
				try:
					data = stream.read()
				finally:
					stream.close()
				
				found = re.findall('youtube\.com/v/([-A-Za-z0-9_]+)', data)
				
				if not found:
					found = re.findall('video_id=([-A-Za-z0-9_]+)', data)
				
				if not found:
					found = re.findall('v=([-A-Za-z0-9_]+)', data)
			
				if not found:
					showError(
						_("Save YouTube Video"),
						_("Can\'t get the youtube-videos ID out of the url or the referred site content.\n"
						  "Maybe you have to be logged in to view this video?"))

				for video_id in found:
					if video_id not in video_ids:
						video_ids.append(video_id)
			
		conf.read([conf_file])

		if conf.has_option('get_yt_video','save_dir'):
			save_dir = conf.get('get_yt_video','save_dir')
		
		else:
			save_dir = os.curdir
		
		if not conf.has_section('get_yt_video'):
			conf.add_section('get_yt_video')
		
		for video_id in video_ids:
			file_name, vid_url = get_filename(save_dir, video_id, format)
				
			while True:
				file_name, serr, code = showSaveFile(_("Save YouTube Video"),file_name)

				if code != 0: # Cancel
					break

				if not os.path.exists(file_name):
					break

				ignore, serr, code = showWarningYesNoCancel(
						_("Save YouTube Video"),
						_('File "%(file)s" already exists.\nDo you want to replace it?') %
						{'file':file_name})
						
				if code == 0: # Yes
					break

				elif code == 1: # No
					continue

				else: # code == 2: Cancel
					break

			if code == 0: # Ok [+ Replace]
				save_dir = os.path.split(file_name)[0]

				try:
					conf.set('get_yt_video','save_dir',save_dir)
					conf.write(open(conf_file,'w'))
				except:
					traceback.print_exc(file=sys.stderr)
			
				try:
					get_video(vid_url, file_name, action)
				
				except Exception, errmsg:
					showError(_('Save YouTube Video'),str(errmsg))

	except Exception, err:
# debug output:
#		try:
#			from cStringIO import StringIO
#		except ImportError:
#			from StringIO import StringIO
#		sio = StringIO()
#		traceback.print_exc(file=sio)
#		showError(_('Save YouTube Video'),"<pre>%s</pre>" % htmlesc(sio.getvalue()))
#		sio.close()
		
		traceback.print_exc(file=sys.stderr)
		showError(_('Save YouTube Video'),str(err))

if __name__ == "__main__":
	main()
